/* pcx2ppm 1.2 - convert pcx to ppm
 * based on zgv's readpcx.c
 * public domain by RJM
 *
 * this is incredibly messy
 *
 * 1999-03-16	updated to support 24-bit.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define _PIC_OK			 0
#define _PICERR_NOFILE		-1
#define _PICERR_NOMEM		-2
#define _PICERR_UNSUPPORTED	-6
#define _PICERR_CORRUPT		-7

/* macro often used to close input file and return an error from above */
#define CLOSE_AND_RET(x)	do { fclose(in); return(x); } while(0)

typedef void (*hffunc)(int,int);

typedef unsigned char byte;


struct pcxhed
  {
  byte manuf,ver,encod,bpp;		/* 0 - 3 gen format */
  byte x1lo,x1hi,y1lo,y1hi;		/* 4 - 11 size */
  byte x2lo,x2hi,y2lo,y2hi;
  byte unused1[4];			/* 12 - 15 scrn size */
  byte pal16[48];			/* 16 - 63 4-bit palette */
  byte reserved;			/* 64 reserved */
  byte nplanes;				/* 65 num of bitplanes */
  byte bytelinelo,bytelinehi;		/* 66 - 67 bytes per line */
  byte unused2[60];			/* 68 - 127 unused */
  };  /* palette info is after image data */


hffunc howfar;

int width,height;
unsigned char *theimage;
unsigned char *pal;

int pixelsize=3;


/* prototypes */
void dispbyte(unsigned char *ptr,int *xp,int *yp,int c,int w,int h,
              int real_bpp,int byteline,int *planep,int *pmaskp);



int read_pcx_file(char *filename,hffunc howfarfunc,unsigned char **pal)
{
FILE *in;
int w,h,bytepp,x,y,yy,byteline,plane,pmask;
unsigned char *ptr;
struct pcxhed header;
int count,waste;
long bytemax,bytesdone;
byte inbyte,inbyte2;
int real_bpp;		/* how many bpp file really is */

theimage=NULL;
*pal=NULL;

if(filename==NULL)
  in=stdin;
else
  if((in=fopen(filename,"rb"))==NULL)
    return(_PICERR_NOFILE);

fread(&header,1,sizeof(struct pcxhed),in);
if(header.manuf!=10 || header.encod!=1)
  CLOSE_AND_RET(_PICERR_UNSUPPORTED);

/* header.bpp=1, header.nplanes=1 = 1-bit.
 * header.bpp=1, header.nplanes=4 = 4-bit.
 * header.bpp=8, header.nplanes=1 = 8-bit.
 * header.bpp=8, header.nplanes=3 = 24-bit.
 * anything else gives an `unsupported' error.
 */
real_bpp=0; bytepp=1;
switch(header.bpp)
  {
  case 1:
    switch(header.nplanes)
      {
      case 1: real_bpp=1; break;
      case 4: real_bpp=4; break;
      }
    break;
  
  case 8:
    switch(header.nplanes)
      {
      case 1: real_bpp=8; break;
      case 3: real_bpp=24; bytepp=3; break;
      }
    break;
  }

if(!real_bpp)
  CLOSE_AND_RET(_PICERR_UNSUPPORTED);

if((*pal=calloc(768,1))==NULL)
  CLOSE_AND_RET(_PICERR_NOMEM);

w=(header.x2lo+256*header.x2hi)-(header.x1lo+256*header.x1hi)+1;
h=(header.y2lo+256*header.y2hi)-(header.y1lo+256*header.y1hi)+1;
byteline=header.bytelinelo+256*header.bytelinehi;

if(w==0 || h==0)
  CLOSE_AND_RET(_PICERR_CORRUPT);

x=0; y=0;
bytemax=w*h;
if(real_bpp==1 || real_bpp==4)
  bytemax=(1<<30);	/* we use a 'y<h' test instead for these files */

if((theimage=malloc(w*(h+2)*bytepp))==NULL)
  CLOSE_AND_RET(_PICERR_NOMEM);

/* need this if more than one bitplane */
memset(theimage,0,w*h*bytepp);

bytesdone=0;
ptr=theimage;

/* start reading image */
for(yy=0;yy<h;yy++)
  {
  if(howfarfunc!=NULL) howfarfunc(y,h);
  
  plane=0; pmask=1;
  
  y=yy;
  x=0;
  while(y==yy)
    {
    inbyte=fgetc(in);
    if(inbyte<192)
      {
      dispbyte(ptr,&x,&y,inbyte,w,h,real_bpp,
      	byteline,&plane,&pmask);
      bytesdone++;
      }
    else
      {
      inbyte2=fgetc(in);
      inbyte&=63;
      for(count=0;count<inbyte;count++)
        dispbyte(ptr,&x,&y,inbyte2,w,h,real_bpp,
        	byteline,&plane,&pmask);
      bytesdone+=inbyte;
      }
    }
  }

/* read palette */
switch(real_bpp)
  {
  case 1:
    ptr=*pal;
    ptr[0]=ptr[256]=ptr[512]=0;
    ptr[1]=ptr[257]=ptr[513]=255;
    break;
  
  case 4:
    /* 4-bit, palette is embedded in header */
    ptr=*pal;
    for(x=0;x<16;x++)
      {
      ptr[  0]=header.pal16[x*3  ];
      ptr[256]=header.pal16[x*3+1];
      ptr[512]=header.pal16[x*3+2];
      ptr++;
      }
    break;
  
  case 8:
    /* 8-bit */
    waste=fgetc(in);                    /* ditch splitter byte */
    /* palette already allocated */
    ptr=*pal;
    for(x=0;x<256;x++)
      {
      ptr[0]  =fgetc(in);
      ptr[256]=fgetc(in);
      ptr[512]=fgetc(in);
      ptr++;
      }
    break;
  
  case 24:
    /* no palette */
    break;
  }

width=w;
height=h;
pixelsize=bytepp;

if(in!=stdin) fclose(in);
return(_PIC_OK);  
}


void dispbyte(unsigned char *ptr,int *xp,int *yp,int c,int w,int h,
              int real_bpp,int byteline,int *planep,int *pmaskp)
{
switch(real_bpp)
  {
  case 1: case 4:
    {
    /* mono or 4-bit */
    int f;
    unsigned char *dstptr;
    
    if((*yp)>=h) return;
    
    dstptr=ptr+(*yp)*w+*xp;
    w=byteline*8;
    
    for(f=0;f<8;f++)
      {
      *dstptr++|=(c&(0x80>>(f&7)))?(*pmaskp):0;
      (*xp)++;
      if(*xp>=w)
        {
        if(real_bpp==1) { (*xp)=0,(*yp)++; return; }
        /* otherwise, it's 4 bpp */
        (*xp)=0;
        (*planep)++; (*pmaskp)<<=1;
        if(*planep==4) { (*yp)++; return; }
        }
      if((*yp)>=h) return;
      }
    }
    break;
  
  case 8:
    *(ptr+(*yp)*w+*xp)=c;
    (*xp)++; if(*xp>=w) (*xp)=0,(*yp)++;
    break;
  
  case 24:
    *(ptr+((*yp)*w+*xp)*3+(2-(*planep)))=c;
    (*xp)++;
    if(*xp>=w)
      {
      (*xp)=0;
      (*planep)++; /* no need to change pmask */
      if(*planep==3) { (*yp)++; return; }
      }
    break;
  }
}



int main(int argc,char *argv[])
{
unsigned char *pal;
int x,y,c;

if(argc>2)
  {
  fprintf(stderr,"usage: pcx2ppm [input.pcx] >output.ppm\n");
  exit(1);
  }

if(read_pcx_file((argc==1)?NULL:argv[1],NULL,&pal)!=_PIC_OK)
  {
  fprintf(stderr,"pcx2ppm: error reading image\n");
  exit(1);
  }

/* otherwise, output image */
printf("P6\n%d %d\n255\n",width,height);

for(y=0;y<height;y++)
  for(x=0;x<width;x++)
    if(pixelsize==1)
      {
      /* lookup index in palette */
      c=theimage[y*width+x];
      putchar(pal[  0+c]);
      putchar(pal[256+c]);
      putchar(pal[512+c]);
      }
    else
      {
      /* just copy */
      putchar(theimage[(y*width+x)*3+2]);
      putchar(theimage[(y*width+x)*3+1]);
      putchar(theimage[(y*width+x)*3  ]);
      }

exit(0);
}
